/*
 * HFS/2 - A Hierarchical File System Driver for OS/2
 * Copyright (C) 1996, 1997 Marcus Better
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* r0comm.c
   Ring 0 side of communications
*/

#define INCL_DOSERRORS
#define INCL_DOSMISC
#define INCL_NOPMAPI
#include <os2.h>

#include <dhcalls.h>

#include <string.h>
#include <ctype.h>
#include <dos.h>

#include "fsd.h"
#include "..\r0r3shar.h"
#include "fsh.h"
#include "r0.h"
#include "..\banner.h"
#include "..\hfsctl.h"

#ifdef DEBUG
#include "log.h"
#endif

/* Local prototypes */
static void DetachCP(void);
static int parse_args(char *);
static void queue_remove(int);
static int get_new_request(void);
static void queue_append(int);
static void free_request(int);
static int find_request(unsigned long);

/* Flag set if control program currently attached */
int CPAttached;

/* Pointer to device helper entry point */
PFN Device_Help;

/* Process ID of the control program */
static PID32 CPPID;

/* Options */
static int silent;

/* Semaphore that signals when a request has arrived */
static unsigned long req_waiting;

/* Timeout for requests in milliseconds */
static unsigned long request_timeout = 60*1000L;
static unsigned long extra_timeout = 180*1000L;

/* The request queue. First element is sentinel. */
#define REQ_QUEUE_SIZE 16
static FS_REQUEST req_queue[REQ_QUEUE_SIZE];
static unsigned req_free_head;

/* Request ID counter */
static unsigned long current_req_id;

#define DATA_BUF_SIZE (64*1024L)

/* The list of mounted volumes */
#define MAX_DRIVE            25
#define MAX_MOUNTSTRUCS      50

typedef struct MountStruc {
  unsigned short drive;
  unsigned short status;
  unsigned short hVPB;
} MOUNTSTRUC;

static MOUNTSTRUC MountList[MAX_MOUNTSTRUCS];


/******************************************************************************
**
** FS_INIT - Initialize the IFS
**
** Parameters
** ---------
** char far *szParm                    pointer to command line parameters
** unsigned long pDevHlp               pointer to DevHlp entry point
** unsigned long far *pMiniFSD         pointer to data passed between the
**                                     mini-FSD and the IFS
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds FS_INIT(char far *szParm,
                                             unsigned long pDevHlp,
                                             unsigned long far *pMiniFSD)
{
  int i;
  USHORT rc;
  SEL GDTSels[2];
  ULONG dummy;

  Startup();
  Device_Help = (PFN)pDevHlp;

  if(szParm != NULL && parse_args(szParm))
    return ERROR_BAD_ARGUMENTS;

  if(!silent)
    DosPutMessage(1, sizeof(banner)-1, banner);

  /* Allocate GDT selectors */
  for (i=1; i < REQ_QUEUE_SIZE; i++) {
    rc = DevHelp_AllocGDTSelector(GDTSels, 2);
    if (rc)
      return rc;
    /* Make far pointers from the selectors */
    req_queue[i].data_buf = MK_FP(GDTSels[0], 0);
    req_queue[i].parm_buf = MK_FP(GDTSels[1], 0);
  }

  /* Initialize request queue by placing all elements in free list. */
  req_free_head = 1;
  for (i = 1; i < REQ_QUEUE_SIZE-1; i++)
    req_queue[i].next = i+1;
  req_queue[REQ_QUEUE_SIZE-1].next = 0;

  return NO_ERROR;
}

/******************************************************************************
**
** FS_SHUTDOWN - Prepare for a system shutdown
**
** Parameters
** ---------
** unsigned short usType               flavor of call
**   values:
**     SD_BEGIN                  beginning the shutdown process
**     SD_COMPLETE               finishing the shutdown process
** unsigned long ulReserved            reserved
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds FS_SHUTDOWN(unsigned short usType,
                                                 unsigned long ulReserved)
{
  return NO_ERROR;
}

/******************************************************************************
**
** FS_FSCTL - Extended IFS control
**
** Parameters
** ---------
** union argdat far *pArgdat
** unsigned short iArgType             argument type
**   values:
**     FSCTL_ARG_FILEINSTANCE    filehandle directed
**     FSCTL_ARG_CURDIR          pathname directed
**     FSCTL_ARG_NULL            FSD directed
** unsigned short func                 function code
**   values:
**     FSCTL_FUNC_NONE           NULL function
**     FSCTL_FUNC_NEW_INFO       return error code information
**     FSCTL_FUNC_EASIZE         return max EA size and max EA list size
** char far *pParm                     UNVERIFIED pointer to parameter area
** unsigned short lenParm              size of area pointed to by pParm
** unsigned short far *plenParmOut     length of parameters passed in pParm
** char far *pData                     UNVERIFIED pointer to information area
** unsigned short lenData              size of area pointed to by pData
** unsigned short far *plenDataOut     length of parameters passed in pData
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds FS_FSCTL(union argdat far *pArgdat,
                                              unsigned short iArgType,
                                              unsigned short func,
                                              char far *pParm,
                                              unsigned short lenParm,
                                              unsigned short far *plenParmOut,
                                              char far *pData,
                                              unsigned short lenData,
                                              unsigned short far *plenDataOut)
{
  int rc;

  switch (func) {
  case FSCTL_FUNC_NONE:
    return NO_ERROR;

  case FSCTL_FUNC_NEW_INFO:
  case FSCTL_FUNC_EASIZE:
    return ERROR_NOT_SUPPORTED;

  case FSCTL_FUNC_READSECTORS: {
    struct vpfsi *pvpfsi;
    struct vpfsd *pvpfsd;
    unsigned short sector_size;

    /* Verify parameter and data areas */
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(RWSTRUC));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, &((RWSTRUC *)pParm)->n_sectors,
		      sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;

    rc=FSH_GETVOLPARM(((RWSTRUC *)pParm)->hVPB, &pvpfsi, &pvpfsd);
    if(rc!=NO_ERROR)
      return rc;
    sector_size = pvpfsi->vpi_bsize;

    rc = FSH_PROBEBUF(PB_OPWRITE, pData,
		      ((RWSTRUC *)pParm)->n_sectors * sector_size);
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenDataOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    /* Verify buffer size */
    if( lenData < ((RWSTRUC *)pParm)->n_sectors * sector_size ) {
      *plenDataOut = ((RWSTRUC *)pParm)->n_sectors * sector_size;
      return ERROR_BUFFER_OVERFLOW;
    }

    /* Read sectors */
    rc = FSH_DOVOLIO(DVIO_OPREAD,
		     DVIO_ALLFAIL|DVIO_ALLRETRY|DVIO_ALLABORT,
		     ((RWSTRUC *)pParm)->hVPB,
		     pData,
		     &((RWSTRUC *)pParm)->n_sectors,
		     ((RWSTRUC *)pParm)->starting_sector);
    *plenParmOut = sizeof(unsigned short);
    *plenDataOut = ((RWSTRUC *)pParm)->n_sectors * sector_size;
    return rc;
  }

  case FSCTL_FUNC_WRITESECTORS: {
    struct vpfsi *pvpfsi;
    struct vpfsd *pvpfsd;
    unsigned short sector_size;

    /* Verify parameter and data areas */
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(RWSTRUC));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, &((RWSTRUC *)pParm)->n_sectors,
		      sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;

    rc=FSH_GETVOLPARM(((RWSTRUC *)pParm)->hVPB, &pvpfsi, &pvpfsd);
    if(rc!=NO_ERROR)
      return rc;
    sector_size = pvpfsi->vpi_bsize;

    rc = FSH_PROBEBUF(PB_OPREAD, pData,
		      ((RWSTRUC *)pParm)->n_sectors * sector_size);
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenDataOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    /* Verify buffer size */
    if( lenData < ((RWSTRUC *)pParm)->n_sectors * sector_size ) {
      *plenDataOut = ((RWSTRUC *)pParm)->n_sectors * sector_size;
      return ERROR_BUFFER_OVERFLOW;
    }

    /* Write sectors */
    rc = FSH_DOVOLIO(DVIO_OPWRITE,
		     DVIO_ALLFAIL|DVIO_ALLRETRY|DVIO_ALLABORT,
		     ((RWSTRUC *)pParm)->hVPB,
		     pData,
		     &((RWSTRUC *)pParm)->n_sectors,
		     ((RWSTRUC *)pParm)->starting_sector);
    *plenParmOut = sizeof(unsigned short);
    *plenDataOut = ((RWSTRUC *)pParm)->n_sectors * sector_size;
    return rc;
  }

  case FSCTL_FUNC_WRITESECTORS_NOHARDERR: {
    /* Write without hard error signaling */
    struct vpfsi *pvpfsi;
    struct vpfsd *pvpfsd;
    unsigned short sector_size;

    /* Verify parameter and data areas */
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(RWSTRUC));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, &((RWSTRUC *)pParm)->n_sectors,
		      sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;

    rc=FSH_GETVOLPARM(((RWSTRUC *)pParm)->hVPB, &pvpfsi, &pvpfsd);
    if(rc!=NO_ERROR)
      return rc;
    sector_size = pvpfsi->vpi_bsize;

    rc = FSH_PROBEBUF(PB_OPREAD, pData,
		      ((RWSTRUC *)pParm)->n_sectors * sector_size);
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenDataOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    /* Verify buffer size */
    if( lenData < ((RWSTRUC *)pParm)->n_sectors * sector_size ) {
      *plenDataOut = ((RWSTRUC *)pParm)->n_sectors * sector_size;
      return ERROR_BUFFER_OVERFLOW;
    }

    /* Write sectors */
    rc = FSH_DOVOLIO(DVIO_OPWRITE | DVIO_OPHARDERR,
		     DVIO_ALLFAIL|DVIO_ALLRETRY|DVIO_ALLABORT,
		     ((RWSTRUC *)pParm)->hVPB,
		     pData,
		     &((RWSTRUC *)pParm)->n_sectors,
		     ((RWSTRUC *)pParm)->starting_sector);
    *plenParmOut = sizeof(unsigned short);
    *plenDataOut = ((RWSTRUC *)pParm)->n_sectors * sector_size;
    return rc;
  }

  case FSCTL_FUNC_FINDDUPHVPB: {
    unsigned short oldVPB;

    /* Verify parameter area */
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, pParm, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;

    /* Call FS helper */
    rc = FSH_FINDDUPHVPB(*(unsigned short *)pParm, &oldVPB);
    *(unsigned short *)pParm = oldVPB;
    *plenParmOut = sizeof(unsigned short);
    return rc;
  }

  case FSCTL_FUNC_GET_MTSTAT: {
    PMTSTATSTRUC mts=(PMTSTATSTRUC)pParm;
    unsigned short i;

    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, pParm, sizeof(MTSTATSTRUC));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;

    /* Search the mount list for the given hVPB */
    i=0;
    while(i<MAX_MOUNTSTRUCS && *(unsigned short *)pParm != MountList[i].hVPB)
      i++;
    if(i==MAX_MOUNTSTRUCS)
      return ERROR_INVALID_PARAMETER;

    mts->drive = MountList[i].drive;
    mts->status = MountList[i].status;
    mts->hVPB = MountList[i].hVPB;
    *plenParmOut = sizeof(MTSTATSTRUC);
    return NO_ERROR;
  }

  case FSCTL_FUNC_SET_MTSTAT: {
    PMTSTATSTRUC mts=(PMTSTATSTRUC)pParm;
    unsigned short i;

    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(MTSTATSTRUC));
    if( rc != NO_ERROR )
      return rc;

    i=0;
    while(i<MAX_MOUNTSTRUCS && MountList[i].hVPB!=mts->hVPB)
      i++;
    if(i==MAX_MOUNTSTRUCS) {
      /* Find a free mount struc */
      i=0;
      while(i<MAX_MOUNTSTRUCS && MountList[i].status!=MTSTAT_FREE)
	i++;
      if(i==MAX_MOUNTSTRUCS)
	return ERROR_BUFFER_OVERFLOW;
    }
    MountList[i].hVPB = mts->hVPB;
    MountList[i].drive = mts->drive;
    MountList[i].status = mts->status;
    return NO_ERROR;
  }

  case FSCTL_FUNC_WILDMATCH:
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(WILDSTRUC));
    if( rc != NO_ERROR )
      return rc;

    return FSH_WILDMATCH(((PWILDSTRUC)pParm)->pattern,
			 ((PWILDSTRUC)pParm)->text);

  case FSCTL_FUNC_UPPERCASE:
    rc = FSH_PROBEBUF(PB_OPREAD, pData, lenData);
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, pData, lenData);
    if( rc != NO_ERROR )
      return rc;

    return FSH_UPPERCASE(pData, lenData, pData);

  case FSCTL_FUNC_SETVOLUME:
    /* Verify parameter area */
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, 2*sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_SETVOLUME(*(unsigned short *)pParm, *(unsigned short *)(pParm+2));
    return rc;

  case FSCTL_FUNC_SET_TIMEOUT:
    /* Allows dynamically changing the request time-out.

       Parameter: New timeout value (unsigned short)
       */

    /* Verify parameter area */
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(unsigned long));
    if (rc != NO_ERROR)
      return rc;

    request_timeout = *(unsigned short *)pParm;
    return NO_ERROR;

  case FSCTL_FUNC_INIT:
    if (CPAttached)
      return 0xef00;

    /* Verify size of and addressability of the pParm area */
    if (lenParm != sizeof(INITSTRUC))
      return 0xef01;
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, lenParm);
    if (rc != NO_ERROR)
      return rc;

    /* Save control program's process ID */
    CPPID = ((INITSTRUC *)pParm)->CPPID;

    /* The control program is now attached */
    CPAttached = 1;

#ifdef DEBUG
    printk("Attached to control program, PID = %lu\n", (unsigned long)CPPID);
#endif
    return NO_ERROR;

  case FSCTL_FUNC_GET_REQUEST: {
    /* Called by the control program to fetch the next request. We block
       until a request comes in. Then we set up GDT selectors to the
       control program's parameter and data buffers so that the request
       thread can get at them, and tell it to copy in the data.

       Parameters: OPDATA structure (to be filled in)
       Data: Data buffer for command-dependent data. Must be 64K.
       */

    FS_REQUEST *req;
    LIN linaddr;

    if (!CPAttached)
      return 0xef02;

    /* Verify parameter and data area */
    if (lenParm != sizeof(OPDATA))
      return ERROR_INVALID_PARAMETER;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenDataOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, pParm, sizeof(OPDATA));
    if (rc != NO_ERROR)
      return rc;

    /* Wait for a request */
    if (req_queue[0].next == 0) {
      rc = FSH_SEMSETWAIT(&req_waiting, TO_INFINITE);
      if (rc != NO_ERROR)
	return rc;
    }

    req = &req_queue[req_queue[0].next];
    if (req-req_queue == 0)
      return 0xef03;
    queue_remove(req_queue[0].next);

    req->request_id = ++current_req_id;

    /* Map the GDT selectors to the data and parameter buffers */
    rc = DevHelp_VirtToLin(SELECTOROF(pParm), OFFSETOF(pParm), &linaddr);
    if (rc)
      return 0xef04;
    rc = DevHelp_LinToGDTSelector(SELECTOROF(req->parm_buf), linaddr, 
				  sizeof(OPDATA));
    if (rc)
      return 0xef05;

    if (req->data_len > 0) {
      rc = DevHelp_VirtToLin(SELECTOROF(pData), OFFSETOF(pData), &linaddr);
      if (rc)
	return 0xef06;
      rc = DevHelp_LinToGDTSelector(SELECTOROF(req->data_buf), 
				    linaddr, req->data_len);
      if (rc)
	return rc;
    }

    /* Tell the request thread to fill the buffers */
    FSH_SEMSET(&req->control_sem);
    FSH_SEMCLEAR(&req->request_sem);
    FSH_SEMWAIT(&req->control_sem, TO_INFINITE);
    if (rc != NO_ERROR)
      return rc;

    *plenParmOut = sizeof(OPDATA);
    *plenDataOut = req->data_len;
    return NO_ERROR;
  }

  case FSCTL_FUNC_REQUEST_DONE: {
    /* Called by the control program when a request has been completed
       to return the results. We set up the GDT selectors to the parameters
       and data for the request thread and tell it to transfer the results.

       Parameters: An OPDATA structure.
       Data: Command-dependent data.
       */

    int i;
    FS_REQUEST *req;
    OPDATA *opdata = (OPDATA *)pParm;
    LIN linaddr;

    /* Verify parameter and data area */
    if (lenParm != sizeof(OPDATA))
      return ERROR_INVALID_PARAMETER;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenDataOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(OPDATA));
    if (rc != NO_ERROR)
      return rc;
    
    i = find_request(opdata->request_id);
    if (i == 0)
      return 0xef08;
    req = &req_queue[i];
    
    /* Map the GDT selectors to the data and parameter buffers */
    rc = DevHelp_VirtToLin(SELECTOROF(pParm), OFFSETOF(pParm), &linaddr);
    if (rc)
      return 0xef09;
    rc = DevHelp_LinToGDTSelector(SELECTOROF(req->parm_buf), linaddr, lenParm);
    if (rc)
      return 0xef0a;

    rc = DevHelp_VirtToLin(SELECTOROF(pData), OFFSETOF(pData), &linaddr);
    if (rc)
      return 0xef0b;
    rc = DevHelp_LinToGDTSelector(SELECTOROF(req->data_buf), linaddr, 65536L);
    if (rc)
      return rc;

    /* Tell the request thread to copy the data from the buffers */
    FSH_SEMSET(&req->control_sem);
    FSH_SEMCLEAR(&req->request_sem);
    FSH_SEMWAIT(&req->control_sem, TO_INFINITE);
    if (rc != NO_ERROR)
      return rc;

    free_request(i);

    return NO_ERROR;
  }
  
  case FSCTL_FUNC_PROLONG_REQUEST: {
    /* Called when a request may take a long time, for example if a large
       diskette read needs to be done. We set a flag in the request
       struct to indicate that the request thread should wait longer
       than usual before time out.

       Parameter: The request id (unsigned long)
       */

    int i;
    FS_REQUEST *req;

    /* Verify parameter and data area */
    if (lenParm != sizeof(unsigned long))
      return ERROR_INVALID_PARAMETER;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPWRITE, plenDataOut, sizeof(unsigned short));
    if( rc != NO_ERROR )
      return rc;
    rc = FSH_PROBEBUF(PB_OPREAD, pParm, sizeof(unsigned long));
    if (rc != NO_ERROR)
      return rc;
    
    i = find_request(*(unsigned long *)pParm);
    if (i == 0)
      return 0xef09;
    req_queue[i].prolong_request = 1;
    return NO_ERROR;
  }

  /* User functions */
  case HFS_FUNC_GET_RESFORK_SIZE: {
    OPDATA opdata;
    struct vpfsi *pvpfsi;
    struct vpfsd *pvpfsd;

    if(iArgType != FSCTL_ARG_FILEINSTANCE)
      return ERROR_INVALID_CATEGORY;

    if(FSH_PROBEBUF(PB_OPWRITE, pParm, sizeof(unsigned long)) != NO_ERROR ||
       FSH_PROBEBUF(PB_OPWRITE, plenParmOut, sizeof(unsigned short)) 
       != NO_ERROR)
      return ERROR_INVALID_PARAMETER;

    if(lenParm<sizeof(unsigned long)) {
      *plenParmOut = sizeof(unsigned long);
      return ERROR_BUFFER_OVERFLOW;
    }

    if (!CPAttached)
      return ERROR_NOT_READY;

    opdata.funccode = CPFC_GET_RESFORK_SIZE;
    FSH_GETVOLPARM(pArgdat->sf.psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
    opdata.p.get_resfork_size.vpfsd = *pvpfsd;
    opdata.p.get_resfork_size.sffsi = *pArgdat->sf.psffsi;
    opdata.p.get_resfork_size.sffsd = *pArgdat->sf.psffsd;

    rc = do_request(&opdata, NULL, 0);
    if (rc != NO_ERROR)
      return rc;

    *pvpfsd = opdata.p.get_resfork_size.vpfsd;
    *pArgdat->sf.psffsi = opdata.p.get_resfork_size.sffsi;
    *pArgdat->sf.psffsd = opdata.p.get_resfork_size.sffsd;

    *(unsigned long *)pParm = opdata.p.get_resfork_size.size;
    *plenParmOut = sizeof(unsigned long);

    end_request(opdata.request_id);
    return opdata.rc;
  }

  case HFS_FUNC_READ_RESFORK: {
    OPDATA opdata;
    struct vpfsi *pvpfsi;
    struct vpfsd *pvpfsd;
    PHFS_READ_RESFORK_PARAMS param = (PHFS_READ_RESFORK_PARAMS)pParm;
    char *pdata;

    if(iArgType != FSCTL_ARG_FILEINSTANCE)
      return ERROR_INVALID_CATEGORY;

    if(FSH_PROBEBUF(PB_OPREAD, pParm, 
		    sizeof(HFS_READ_RESFORK_PARAMS)) != NO_ERROR
       || FSH_PROBEBUF(PB_OPWRITE, plenDataOut, 
		       sizeof(unsigned short)) != NO_ERROR
       || FSH_PROBEBUF(PB_OPWRITE, pData, param->bytes) != NO_ERROR)
      return ERROR_INVALID_PARAMETER;

    if(lenParm<sizeof(unsigned long)) {
      *plenParmOut = sizeof(unsigned long);
      return ERROR_BUFFER_OVERFLOW;
    }

    if (!CPAttached)
      return ERROR_NOT_READY;

    opdata.funccode = CPFC_READ_RESFORK;
    FSH_GETVOLPARM(pArgdat->sf.psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
    opdata.p.read_resfork.vpfsd = *pvpfsd;
    opdata.p.read_resfork.sffsi = *pArgdat->sf.psffsi;
    opdata.p.read_resfork.sffsd = *pArgdat->sf.psffsd;
    opdata.p.read_resfork.start = param->start;
    opdata.p.read_resfork.bytes = param->bytes;
    pData = NULL;

    rc = do_request(&opdata, &pdata, 0);
    if (rc != NO_ERROR)
      return rc;

    *pvpfsd = opdata.p.read_resfork.vpfsd;
    *pArgdat->sf.psffsi = opdata.p.read_resfork.sffsi;
    *pArgdat->sf.psffsd = opdata.p.read_resfork.sffsd;

    if(opdata.p.read_resfork.bytes > param->bytes)
      opdata.p.read_resfork.bytes = param->bytes;

    if(rc==NO_ERROR)
      memcpy(pData, pdata, opdata.p.read_resfork.bytes);
    *plenDataOut = opdata.p.read_resfork.bytes;

    end_request(opdata.request_id);
    return opdata.rc;
  }

  case HFS_FUNC_SYNC: {
    OPDATA opdata;
    struct vpfsi *pvpfsi;
    struct vpfsd *pvpfsd;
    int i, drive;

    if(iArgType != FSCTL_ARG_CURDIR)
      return ERROR_INVALID_CATEGORY;

    if (!CPAttached)
      return ERROR_NOT_READY;

    if(pArgdat->cd.pPath[0] < 'A' || pArgdat->cd.pPath[0] > 'Z')
      return ERROR_INVALID_PARAMETER;
    drive = pArgdat->cd.pPath[0]-'A';
    i=0;
    while(i<MAX_MOUNTSTRUCS && (MountList[i].drive!=drive
				|| MountList[i].status!=MTSTAT_MOUNTED))
      i++;
    if(i==MAX_MOUNTSTRUCS)
      return ERROR_NOT_READY;

    opdata.funccode = CPFC_SYNC;
    FSH_GETVOLPARM(MountList[i].hVPB, &pvpfsi, &pvpfsd);
    opdata.p.sync.vpfsi = *pvpfsi;
    opdata.p.sync.vpfsd = *pvpfsd;

    rc = do_request(&opdata, NULL, 0);
    if (rc != NO_ERROR)
      return rc;

    *pvpfsi = opdata.p.sync.vpfsi;
    *pvpfsd = opdata.p.sync.vpfsd;

    end_request(opdata.request_id);
    return opdata.rc;
  }

  case HFS_FUNC_SYNC_ALL: {
    OPDATA opdata;

    if(iArgType != FSCTL_ARG_NULL)
      return ERROR_INVALID_CATEGORY;

    if (!CPAttached)
      return ERROR_NOT_READY;

    opdata.funccode = CPFC_SYNC_ALL;

    rc = do_request(&opdata, NULL, 0);
    if (rc != NO_ERROR)
      return rc;

    end_request(opdata.request_id);
    return opdata.rc;
  }
  
#ifdef DEBUG
  case HFS_FUNC_GETLOGDATA: {
    char far *p;

    /* Wait on the log semaphore */
    if( (rc=FSH_SEMWAIT(&logbuf_data_ready, TO_INFINITE)) == ERROR_INTERRUPT )
      return rc;

    /* Check the user buffer and length field */
    if( (rc=FSH_PROBEBUF(PB_OPWRITE, pData, lenData)) != NO_ERROR )
      return rc;
    if( (rc=FSH_PROBEBUF(PB_OPWRITE, plenDataOut, sizeof(unsigned short))) != NO_ERROR )
      return rc;

    /* If no data present, set the semaphore */
    if( logbuf_head == logbuf_tail ) {
      *plenDataOut = 0;
      if((rc=FSH_SEMSET(&logbuf_data_ready)) == ERROR_INTERRUPT)
	return rc;
      return NO_ERROR;
    }

    /* Copy data */
    p = pData;
    for( ; logbuf_head!=logbuf_tail && p<pData+lenData; 
	 logbuf_head=(logbuf_head+1)%LOGBUF_SIZE )
      *p++ = logbuf[logbuf_head];
    *plenDataOut = p-pData;

    /* Indicate that the log buffer is empty again */
    FSH_SEMSET(&logbuf_data_ready);

    return NO_ERROR;
  }
#endif

  default:
    return ERROR_NOT_SUPPORTED;
  }
}

/******************************************************************************
**
** FS_EXIT - Notify FSD that a process is ending
**
** Parameters
** ---------
** unsigned short uid                   user ID of process
** unsigned short pid                   process ID of process
** unsigned short pdb                   DOS mode process ID of process
**
******************************************************************************/

#pragma argsused
void far pascal _export _loadds FS_EXIT(unsigned short uid, unsigned short pid,
                                        unsigned short pdb)
{
  /* If the control program is exiting, detach ourselves from it */
  if (CPAttached && pid == CPPID)
    DetachCP();
}

/* Detach from control program */
static void DetachCP(void) {
#ifdef DEBUG
  printk("Detaching from CP\n");
#endif
  CPAttached = 0;
}

/* Parse command line */
static int parse_args(char *args)
{
  char *p;
  int rc=0;

  p = args;
  while(*p) {
    while(*p && isspace(*p))
      p++;
    if(!*p)
      break;

    switch(*p) {
    case '/':
      p++;
      if(!*p) {
        rc=1;
	break;
      }
      switch(toupper(*p)) {
      case 'Q':
	p++;
	silent=1;
	break;
      default:
        rc=1;
      }
      break;
    default:
      rc=1;
    }

    if(rc) break;

    if(*p && !isspace(*p))
      return 1;
  }
  return rc;
}

/* Remove a request from the queue (without adding it to the free list. */
static void queue_remove(int i)
{
  req_queue[req_queue[i].prev].next = req_queue[i].next;
  req_queue[req_queue[i].next].prev = req_queue[i].prev;
}

/* Get a free queue element */
static int get_new_request(void)
{
  if (req_free_head) {
    int i = req_free_head;
    req_free_head = req_queue[i].next;
    return i;
  }
  else
    return 0;
}

/* Add a request to the free list */
static void free_request(int i)
{
  req_queue[i].next = req_free_head;
  req_free_head = i;
  req_queue[i].request_id = 0;
}

/* Append a request to the request list */
static void queue_append(int i)
{
  int last = req_queue[0].prev;
  req_queue[last].next = i;
  req_queue[0].prev = i;
  req_queue[i].next = 0;
  req_queue[i].prev = last;
}
    
/* Find request with given id, not necessarily in the queue. Requests
   are removed from the queue when they are being processed, but are
   not added to the free list until the processing is finished. */
static int find_request(unsigned long request_id)
{
  int i;
  for (i=1; i < REQ_QUEUE_SIZE; i++)
    if (req_queue[i].request_id == request_id)
      break;
  if (i == REQ_QUEUE_SIZE)
    return 0;
  return i;
}

/* Enqueue a request and wait for it to be serviced.
   Parameters:
     opdata        INPUT   the operation parameters
     data          INPUT   pointer to pointer to data buffer
                   OUTPUT  pointer to pointer to returned data
     data_len      INPUT   length of data
   The caller knows the length of the data on output.
*/
int do_request(OPDATA *opdata, char **data, unsigned short data_len)
{
  int rc, i;
  FS_REQUEST *req;
  unsigned long timeout;

  /* The steps for processing the request are as follows:

     1. Fill in the request structure and enqueue it.
     2. Tell the control thread that a request has arrived, and
        wait until it is time for passing the request to the control
	program.
     3. When we unblock, the GDT pointers in the request structures point
        to the control program's parameter and data buffers. We fill them
	in and tell the control thread to return with the request.
     4. Block until the request has been carried out.
     5. Return and let the caller copy the returned data out of the buffers.
     */

#ifdef DEBUG
  printk("Request, code=%hx, datalen=%hu\n", 
	 opdata->funccode,
	 data_len);
#endif

  i = get_new_request();
  if (i == 0)
    return ERROR_NOT_READY;
  req = &req_queue[i];
  queue_append(i);
  req->data_len = data_len;

  /* Now tell the control program a request is waiting, and block */
  FSH_SEMSET(&req->request_sem);
  FSH_SEMCLEAR(&req_waiting);

  rc = FSH_SEMWAIT(&req->request_sem, request_timeout);
  if (rc != NO_ERROR) {
    queue_remove(i);
    if (rc == ERROR_SEM_TIMEOUT) {
#ifdef DEBUG
      printk("Timeout id=%lu\n", req->request_id);
#endif
      return ERROR_NOT_READY;
    }
    else
      return rc;
  }

  /* Fill the data and parameter buffers */
  opdata->request_id = req->request_id;
  *(OPDATA *)(req->parm_buf) = *opdata;
  if (data != NULL && *data != NULL && data_len > 0)
    memcpy(req->data_buf, *data, data_len);

  /* Wait while the request is carried out */
  FSH_SEMSET(&req->request_sem);
  FSH_SEMCLEAR(&req->control_sem);
  
  timeout = request_timeout;
  req->prolong_request = 0;
  for (;;) {
    rc = FSH_SEMWAIT(&req->request_sem, timeout);
    if (rc == ERROR_SEM_TIMEOUT && req->prolong_request) {
#ifdef DEBUG
      printk("Extra time id=%lu\n", req->request_id);
#endif
      req->prolong_request = 0;
      timeout = extra_timeout;
      continue;
    }
    break;
  }
  if (rc != NO_ERROR) {
    queue_remove(i);
    if (rc == ERROR_SEM_TIMEOUT) {
#ifdef DEBUG
      printk("Timeout id=%lu\n", req->request_id);
#endif
      return ERROR_NOT_READY;
    }
    else
      return rc;
  }
#ifdef DEBUG
  printk("Finished code=%hx\n", opdata->funccode);
#endif  
  *opdata = *(OPDATA *)(req->parm_buf);
  if (data != NULL)
    *data = req->data_buf;

  return NO_ERROR;
}

/* Signal that a serviced request has been processed */
void end_request(unsigned long id)
{
  int i = find_request(id);
  FSH_SEMCLEAR(&req_queue[i].control_sem);
}
