/*
 * HFS/2 - A Hierarchical File System Driver for OS/2
 * Copyright (C) 1996, 1997 Marcus Better
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* r3comm.c
   Ring 3 side of communications
*/

#define INCL_DOSERRORS
#define INCL_DOSMEMMGR
#define INCL_DOSFILEMGR
#define INCL_DOSPROCESS
#define INCL_NOCOMMON
#define INCL_NOPMAPI
#include <os2.h>

#include <stddef.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>

#include "libhfs/hfs.h"
#include "fsd.h"
#include "fsdep.h"
#include "../r0r3shar.h"
#include "../banner.h"
#include "names.h"
#include "userfunc.h"
#include "global.h"
#include "log.h"
#include "fsctl.h"

#define DATABUF_SIZE 64*1024

typedef struct func_data {
  char *buf;
  OPDATA opdata;
} FUNCDATA, *PFUNCDATA;

static int parse_args(int, char **);
static void get_requests(void);
static void process_request(void *);

/* Global - file for log output */
FILE *logfile;

/* Global - filename translation mode */
int filename_translation;

/* Global - File system name */
const char FS_NAME[] = "HFS";

/* Option - sync after every write */
static int sync_always=1;

int main(int argc, char **argv) {
  INITSTRUC initdata;
  unsigned long leninitdata;
  unsigned int rc;
  PTIB ptib;
  PPIB ppib;

  fprintf(stderr, banner);

#ifdef DEBUG
  fprintf(stderr, "Debug version\n");
#endif

  logfile = stdout;

  /* Parse command line */
  if(parse_args(argc, argv)) {
    fprintf(stderr, 
	    "\nUsage: HFS [options]\n"
	    "Options:\n"
	    "  /FN:mode    Select filename translation mode.\n"
	    "              mode must be CP850, 7BIT or 8BIT.\n"
            "  /NOSYNC     Do not sync volumes automatically.\n"
	    );
    return 1;
  }

  /* Get info blocks */
  if( (rc=DosGetInfoBlocks(&ptib, &ppib)) != NO_ERROR ) {
    fprintf(stderr, "DosGetInfoBlocks failed - rc = %hd\n", rc);
    exit(1);
  }

  /* Set up the initialization data to be passed to the IFS */
  initdata.CPPID = ppib->pib_ulpid;

  /* Call the IFS to initialize */
  leninitdata = sizeof(initdata);
  rc = do_FSCtl(NULL, 0, NULL, &initdata, sizeof(initdata), &leninitdata,
                FSCTL_FUNC_INIT);
  if (rc != NO_ERROR) {
    fprintf(stderr, "Failed to connect with IFS. Please make sure that\n"
                    "the control program is not already running, and that\n"
                    "you have the line\n"
                    "        DEVICE=d:\\path\\HFS.IFS\n"
                    "in your CONFIG.SYS file.\n");
    exit(1);
  }

  get_requests();
  return 0;
}

/* Get requests from IFS */
void get_requests(void)
{
  PFUNCDATA funcdata;
  int rc;
  ULONG parm_len, data_len;
  char *buf;

  for(;;) {
    buf = malloc(DATABUF_SIZE);
    if (!buf) {
      fprintf(stderr, "Cannot allocate memory\n");
      sleep(2);
      continue;
    }
    /* It seems that we must access the memory before passing it
       to the IFS. Probably it needs to be committed before we can
       map a GDT descriptor to it. */
    memset(buf, 0, DATABUF_SIZE);

    funcdata = malloc(sizeof(FUNCDATA));
    if (!funcdata) {
      fprintf(stderr, "Cannot allocate memory\n");
      free(buf);
      sleep(2);
      continue;
    }
    memset(funcdata, 0, sizeof(funcdata));
    funcdata->buf = buf;

    parm_len = sizeof(OPDATA);
    data_len = DATABUF_SIZE;
    rc = do_FSCtl(buf, DATABUF_SIZE, &data_len, 
		  &funcdata->opdata, sizeof(OPDATA), &parm_len,
		  FSCTL_FUNC_GET_REQUEST);
    if (rc != NO_ERROR) {
      fprintf(stderr, "Failed to get request from IFS, rc = %d\n", rc);
      free(buf);
      free(funcdata);
      sleep(2);
      continue;
    }
#ifdef DEBUG
    printk("Request id=%d, code=%x\n", funcdata->opdata.request_id,
	   funcdata->opdata.funccode);
#endif

    rc = _beginthread(process_request, NULL, 64*1024, funcdata);
    if (rc == -1) {
      fprintf(stderr, "Cannot start thread\n");
      free(buf);
      free(funcdata);
      continue;
    }
  }
}

/* The service thread */
void process_request(void *arg)
{
  PFUNCDATA funcdata = (PFUNCDATA)arg;
  ULONG data_len, parm_len;
  int rc;

  switch (funcdata->opdata.funccode) {
  case CPFC_ATTACH:
    funcdata->opdata.rc = FSD_ATTACH(funcdata->opdata.p.attach.flag, 
				     funcdata->opdata.p.attach.Dev,
				     &funcdata->opdata.p.attach.vpfsd, 
				     &funcdata->opdata.p.attach.cdfsd,
				     funcdata->buf, 
				     &funcdata->opdata.p.attach.Len);
    break;
    
  case CPFC_MOUNT:
    funcdata->opdata.rc = FS_MOUNT(funcdata->opdata.p.mount.flag, 
				   &funcdata->opdata.p.mount.vpfsi,
				   &funcdata->opdata.p.mount.vpfsd, 
				   funcdata->opdata.p.mount.hVPB,
				   funcdata->opdata.p.mount.vol_descr, 
				   funcdata->buf);
    break;
    
  case CPFC_CANCELLOCKREQUEST:
    funcdata->opdata.rc = 
      FS_CANCELLOCKREQUEST(&funcdata->opdata.p.cancellockrequest.sffsi,
			   &funcdata->opdata.p.cancellockrequest.sffsd,
			   &funcdata->opdata.p.cancellockrequest.LockRange);
    break;
    
  case CPFC_CHDIR:
    funcdata->opdata.rc = FS_CHDIR(&funcdata->opdata.p.chdir.vpfsd, 
				   funcdata->opdata.p.chdir.flag,
				   &funcdata->opdata.p.chdir.cdfsi,
				   &funcdata->opdata.p.chdir.cdfsd,
				   funcdata->opdata.p.chdir.Dir,
				   funcdata->opdata.p.chdir.iCurDirEnd);
    break;
    
  case CPFC_CHGFILEPTR:
    funcdata->opdata.rc = FS_CHGFILEPTR(&funcdata->opdata.p.chgfileptr.vpfsd,
					&funcdata->opdata.p.chgfileptr.sffsi,
					&funcdata->opdata.p.chgfileptr.sffsd,
					funcdata->opdata.p.chgfileptr.offset,
					funcdata->opdata.p.chgfileptr.type,
					funcdata->opdata.p.chgfileptr.IOflag);
    break;
    
  case CPFC_CLOSE:
    funcdata->opdata.rc = FS_CLOSE(&funcdata->opdata.p.close.vpfsd,
				   funcdata->opdata.p.close.type,
				   funcdata->opdata.p.close.IOflag,
				   &funcdata->opdata.p.close.sffsi,
				   &funcdata->opdata.p.close.sffsd);
    break;
    
  case CPFC_COMMIT:
    funcdata->opdata.rc = FS_COMMIT(&funcdata->opdata.p.commit.vpfsd,
				    funcdata->opdata.p.commit.type,
				    funcdata->opdata.p.commit.IOflag,
				    &funcdata->opdata.p.commit.sffsi,
				    &funcdata->opdata.p.commit.sffsd);
    break;
    
  case CPFC_COPY:
    funcdata->opdata.rc = FS_COPY(&funcdata->opdata.p.copy.vpfsd,
				  funcdata->opdata.p.copy.flag,
				  &funcdata->opdata.p.copy.cdfsi,
				  &funcdata->opdata.p.copy.cdfsd,
				  funcdata->opdata.p.copy.Src,
				  funcdata->opdata.p.copy.iSrcCurDirEnd,
				  funcdata->opdata.p.copy.Dst,
				  funcdata->opdata.p.copy.iDstCurDirEnd,
				  funcdata->opdata.p.copy.nameType);
    break;
    
  case CPFC_DELETE:
    funcdata->opdata.rc = FS_DELETE(&funcdata->opdata.p.fsdelete.vpfsd,
				    &funcdata->opdata.p.fsdelete.cdfsi,
				    &funcdata->opdata.p.fsdelete.cdfsd,
				    funcdata->opdata.p.fsdelete.File,
				    funcdata->opdata.p.fsdelete.iCurDirEnd);
    break;
    
  case CPFC_FILEATTRIBUTE:
    funcdata->opdata.rc = 
      FS_FILEATTRIBUTE(&funcdata->opdata.p.fileattribute.vpfsd,
		       funcdata->opdata.p.fileattribute.flag,
		       &funcdata->opdata.p.fileattribute.cdfsi,
		       &funcdata->opdata.p.fileattribute.cdfsd,
		       funcdata->opdata.p.fileattribute.Name,
		       funcdata->opdata.p.fileattribute.iCurDirEnd,
		       &funcdata->opdata.p.fileattribute.Attr);
    break;
    
  case CPFC_FILEINFO:
    funcdata->opdata.rc = FS_FILEINFO(&funcdata->opdata.p.fileinfo.vpfsd,
				      funcdata->opdata.p.fileinfo.flag,
				      &funcdata->opdata.p.fileinfo.sffsi,
				      &funcdata->opdata.p.fileinfo.sffsd,
				      funcdata->opdata.p.fileinfo.level,
				      funcdata->buf,
				      funcdata->opdata.p.fileinfo.cbData,
				      funcdata->opdata.p.fileinfo.IOflag,
				      &funcdata->opdata.p.fileinfo.oError);
    break;
    
  case CPFC_FILELOCKS:
    funcdata->opdata.rc = 
      FS_FILELOCKS(&funcdata->opdata.p.filelocks.sffsi,
		   &funcdata->opdata.p.filelocks.sffsd,
		   &funcdata->opdata.p.filelocks.UnLockRange,
		   &funcdata->opdata.p.filelocks.LockRange,
		   funcdata->opdata.p.filelocks.timeout,
		   funcdata->opdata.p.filelocks.flags);
    break;
    
  case CPFC_FINDCLOSE:
    funcdata->opdata.rc = FS_FINDCLOSE(&funcdata->opdata.p.findclose.vpfsd,
				       &funcdata->opdata.p.findclose.fsfsi,
				       &funcdata->opdata.p.findclose.fsfsd);
    break;
    
  case CPFC_FINDFIRST:
    funcdata->opdata.rc = FS_FINDFIRST(&funcdata->opdata.p.findfirst.vpfsd,
				       &funcdata->opdata.p.findfirst.cdfsi,
				       &funcdata->opdata.p.findfirst.cdfsd,
				       funcdata->opdata.p.findfirst.Name,
				       funcdata->opdata.p.findfirst.iCurDirEnd,
				       funcdata->opdata.p.findfirst.attr,
				       &funcdata->opdata.p.findfirst.fsfsi,
				       &funcdata->opdata.p.findfirst.fsfsd,
				       funcdata->buf,
				       funcdata->opdata.p.findfirst.cbData,
				       &funcdata->opdata.p.findfirst.cMatch,
				       funcdata->opdata.p.findfirst.level,
				       funcdata->opdata.p.findfirst.flags,
				       &funcdata->opdata.p.findfirst.oError);
    break;
    
  case CPFC_FINDFROMNAME:
    funcdata->opdata.rc = 
      FS_FINDFROMNAME(&funcdata->opdata.p.findfromname.vpfsd,
		      &funcdata->opdata.p.findfromname.fsfsi,
		      &funcdata->opdata.p.findfromname.fsfsd,
		      funcdata->buf,
		      funcdata->opdata.p.findfromname.cbData,
		      &funcdata->opdata.p.findfromname.cMatch,
		      funcdata->opdata.p.findfromname.level,
		      funcdata->opdata.p.findfromname.position,
		      funcdata->opdata.p.findfromname.Name,
		      funcdata->opdata.p.findfromname.flags,
		      &funcdata->opdata.p.findfirst.oError);
    break;
    
  case CPFC_FINDNEXT:
    funcdata->opdata.rc = FS_FINDNEXT(&funcdata->opdata.p.findnext.vpfsd,
				      &funcdata->opdata.p.findnext.fsfsi,
				      &funcdata->opdata.p.findnext.fsfsd,
				      funcdata->buf,
				      funcdata->opdata.p.findnext.cbData,
				      &funcdata->opdata.p.findnext.cMatch,
				      funcdata->opdata.p.findnext.level,
				      funcdata->opdata.p.findnext.flags,
				      &funcdata->opdata.p.findfirst.oError);
    break;
    
  case CPFC_FLUSHBUF:
    funcdata->opdata.rc = FS_FLUSHBUF(&funcdata->opdata.p.flushbuf.vpfsd,
				      funcdata->opdata.p.flushbuf.flag);
    break;
    
  case CPFC_FSINFO:
    funcdata->opdata.rc = FS_FSINFO(funcdata->opdata.p.fsinfo.flag,
				    funcdata->opdata.p.fsinfo.hVPB,
				    &funcdata->opdata.p.fsinfo.vpfsi,
				    &funcdata->opdata.p.fsinfo.vpfsd,
				    funcdata->buf,
				    funcdata->opdata.p.fsinfo.cbData,
				    funcdata->opdata.p.fsinfo.level);
    break;
    
  case CPFC_MKDIR:
    funcdata->opdata.rc = FS_MKDIR(&funcdata->opdata.p.mkdir.vpfsd,
				   &funcdata->opdata.p.mkdir.cdfsi,
				   &funcdata->opdata.p.mkdir.cdfsd,
				   funcdata->opdata.p.mkdir.Name,
				   funcdata->opdata.p.mkdir.iCurDirEnd,
				   funcdata->buf,
				   funcdata->opdata.p.mkdir.flags,
				   &funcdata->opdata.p.mkdir.oError);
    break;
    
  case CPFC_MOVE:
    funcdata->opdata.rc = FS_MOVE(&funcdata->opdata.p.move.vpfsd,
				  &funcdata->opdata.p.move.cdfsi,
				  &funcdata->opdata.p.move.cdfsd,
				  funcdata->opdata.p.move.Src,
				  funcdata->opdata.p.move.iSrcCurDirEnd,
				  funcdata->opdata.p.move.Dst,
				  funcdata->opdata.p.move.iDstCurDirEnd,
				  funcdata->opdata.p.move.flags);
    break;
    
  case CPFC_NEWSIZE:
    funcdata->opdata.rc = FS_NEWSIZE(&funcdata->opdata.p.newsize.vpfsd,
				     &funcdata->opdata.p.newsize.sffsi,
				     &funcdata->opdata.p.newsize.sffsd,
				     funcdata->opdata.p.newsize.len,
				     funcdata->opdata.p.newsize.IOflag);
    break;
    
  case CPFC_OPENCREATE:
    funcdata->opdata.rc = 
      FS_OPENCREATE(&funcdata->opdata.p.opencreate.vpfsd,
		    &funcdata->opdata.p.opencreate.cdfsi,
		    &funcdata->opdata.p.opencreate.cdfsd,
		    funcdata->opdata.p.opencreate.Name,
		    funcdata->opdata.p.opencreate.iCurDirEnd,
		    &funcdata->opdata.p.opencreate.sffsi,
		    &funcdata->opdata.p.opencreate.sffsd,
		    funcdata->opdata.p.opencreate.ulOpenMode,
		    funcdata->opdata.p.opencreate.usOpenFlag,
		    &funcdata->opdata.p.opencreate.usAction,
		    funcdata->opdata.p.opencreate.usAttr,
		    funcdata->buf,
		    &funcdata->opdata.p.opencreate.fgenflag,
		    &funcdata->opdata.p.opencreate.oError);
    break;
    
  case CPFC_PATHINFO:
    funcdata->opdata.rc = FS_PATHINFO(&funcdata->opdata.p.pathinfo.vpfsd,
				      funcdata->opdata.p.pathinfo.flag,
				      &funcdata->opdata.p.pathinfo.cdfsi,
				      &funcdata->opdata.p.pathinfo.cdfsd,
				      funcdata->opdata.p.pathinfo.Name,
				      funcdata->opdata.p.pathinfo.iCurDirEnd,
				      funcdata->opdata.p.pathinfo.level,
				      funcdata->buf,
				      funcdata->opdata.p.pathinfo.cbData,
				      &funcdata->opdata.p.pathinfo.oError);
    break;
    
  case CPFC_PROCESSNAME:
    funcdata->opdata.rc = 
      FS_PROCESSNAME(funcdata->opdata.p.processname.NameBuf);
    break;
    
  case CPFC_READ:
    funcdata->opdata.rc = FS_READ(&funcdata->opdata.p.read.vpfsd,
				  &funcdata->opdata.p.read.sffsi,
				  &funcdata->opdata.p.read.sffsd,
				  funcdata->buf,
				  &funcdata->opdata.p.read.Len,
				  funcdata->opdata.p.read.IOflag,
				  funcdata->opdata.request_id);
    break;
    
  case CPFC_RMDIR:
    funcdata->opdata.rc = FS_RMDIR(&funcdata->opdata.p.rmdir.vpfsd,
				   &funcdata->opdata.p.rmdir.cdfsi,
				   &funcdata->opdata.p.rmdir.cdfsd,
				   funcdata->opdata.p.rmdir.Name,
				   funcdata->opdata.p.rmdir.iCurDirEnd);
    break;
    
  case CPFC_VERIFYUNCNAME:
    funcdata->opdata.rc = 
      FS_VERIFYUNCNAME(funcdata->opdata.p.verifyuncname.flag,
		       funcdata->opdata.p.verifyuncname.Name);
    break;
    
  case CPFC_WRITE:
    funcdata->opdata.rc = FS_WRITE(&funcdata->opdata.p.write.vpfsd,
				   &funcdata->opdata.p.write.sffsi,
				   &funcdata->opdata.p.write.sffsd,
				   funcdata->buf,
				   &funcdata->opdata.p.write.Len,
				   funcdata->opdata.p.write.IOflag,
				   funcdata->opdata.request_id);
    break;
    
    /* User functions */
  case CPFC_GET_RESFORK_SIZE:
    funcdata->opdata.rc = 
      hfs_resfork_size(&funcdata->opdata.p.get_resfork_size.sffsi,
		       &funcdata->opdata.p.get_resfork_size.sffsd,
		       &funcdata->opdata.p.get_resfork_size.size);
    break;
    
  case CPFC_READ_RESFORK:
    funcdata->opdata.rc = 
      hfs_read_resfork(&funcdata->opdata.p.read_resfork.vpfsd,
		       &funcdata->opdata.p.read_resfork.sffsi,
		       &funcdata->opdata.p.read_resfork.sffsd,
		       funcdata->opdata.p.read_resfork.start,
		       &funcdata->opdata.p.read_resfork.bytes,
		       funcdata->buf);
    break;
    
  case CPFC_SYNC:
      funcdata->opdata.rc = hfs_sync(&funcdata->opdata.p.sync.vpfsd);
      break;
      
  case CPFC_SYNC_ALL:
    funcdata->opdata.rc = hfs_sync_all();
    break;
    
  default:
#ifdef DEBUG
    printk("Unknown IFS function code: %d\n", funcdata->opdata.funccode);
#endif
    free(funcdata->buf);
    free(funcdata);
    _endthread();
  }
    
  /* Flush all volumes if /SYNC option was given */
  if(sync_always)
    hfs_flushall();
  
  /* Return results to IFS */
  data_len = DATABUF_SIZE;
  parm_len = sizeof(OPDATA);
  rc = do_FSCtl(funcdata->buf, DATABUF_SIZE, &data_len, 
		&funcdata->opdata, sizeof(OPDATA), &parm_len,
		FSCTL_FUNC_REQUEST_DONE);
  if (rc != NO_ERROR)
    fprintf(stderr, "Failed to return data to IFS, rc=%d\n", rc);

  free(funcdata->buf);
  free(funcdata);
}

static int parse_args(int argc, char *argv[])
{
  int i;
  for(i=1; i<argc; i++) {
    if(argv[i][0]=='/' || argv[i][0]=='-') {
      /* Switch */
      strupr(argv[i]);
      if(strstr(argv[i], "/FN:")==argv[i]) {
	char *p=argv[i]+4;
	if(!strcmp(p, "CP850"))
	  filename_translation=FN_TR_CP850;
	else if(!strcmp(p, "7BIT"))
	  filename_translation=FN_TR_7BIT;
	else if(!strcmp(p, "8BIT"))
	  filename_translation=FN_TR_8BIT;
	else {
	  fprintf(stderr, "Invalid filename translation mode %s\n", p);
	  return 1;
	}
      }
      else if(!strcmp(argv[i], "/NOSYNC")) {
	sync_always = 0;
      }
      else {
        fprintf(stderr, "Invalid option %s\n", argv[i]);
        return 1;
      }
    }
    else {
      /* Unrecognized argument */
      fprintf(stderr, "Invalid argument %s\n", argv[i]);
      return 1;
    }
  }

  return 0;
}
