/*
 * HFS/2 - A Hierarchical File System Driver for OS/2
 * Copyright (C) 1996, 1997 Marcus Better
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#define INCL_NOPMAPI
#define INCL_DOSERRORS
#include <os2.h>

#include <stdio.h>
#include <errno.h>

#include "libhfs.h"
#include "os.h"
#include "os2func.h"

typedef struct {
  HFILE hf;
  unsigned long sectors;
  unsigned short sector_size;
  unsigned long offset;
} VOLINFO, *PVOLINFO;

/*
 * NAME:	os->open()
 * DESCRIPTION:	open a new descriptor from the given path and mode
 */
#pragma argsused
int os_open(void **priv, const char *path, int mode)
{
  PVOLINFO vol;
  HFILE hf;
  unsigned short sector_size;
  unsigned long sectors;

  /* Special case: if path begins with colon, we are doing a FORMAT
     and the file is already opened. The path contains a file handle
     and volume information. */
  if (path[0] == ':') {
    if (sscanf(path, ":%hu:%hu:%lu", &hf, &sector_size, &sectors) != 3)
      ERROR(EACCES, 0);

    vol = os2_malloc(sizeof(VOLINFO));
    vol->hf = hf;
    vol->sectors = sectors;
    vol->sector_size = sector_size;
    vol->offset = 0;

    *priv = (void *)vol;
    return 0;
  }
  else 
    ERROR(EACCES, 0);

fail:
  return -1;
}

/*
 * NAME:	os->same()
 * DESCRIPTION:	return 1 iff path is same as the open descriptor
 */
#pragma argsused
int os_same(void **priv, const char *path)
{
  return 0;
}

/*
 * NAME:	os->lock()
 * DESCRIPTION:	lock a descriptor against concurrent access
 */
#pragma argsused
int os_lock(void **priv, int readonly)
{
  return 0;
}

/*
 * NAME:	os->close()
 * DESCRIPTION:	close an open descriptor
 */
int os_close(void **priv)
{
  free(*priv);
  *priv = NULL;
  return 0;
}

/*
 * NAME:	os->seek()
 * DESCRIPTION:	set a descriptor's seek pointer
 */
long os_seek(void **priv, long offset, int from)
{
  PVOLINFO vol = (PVOLINFO)*priv;
  USHORT method;
  ULONG ofs;

  switch (from) {
  case HFS_SEEK_SET:
    method = FILE_BEGIN;
    break;
  case HFS_SEEK_CUR:
    method = FILE_CURRENT;
    break;
  case HFS_SEEK_END:
    method = FILE_END;
    break;
  default:
    ERROR(EACCES, 0);
  }

  if (DosChgFilePtr(vol->hf, offset, method, &ofs) != NO_ERROR)
    ERROR(EACCES, 0);

  return ofs;

fail:
  return -1;
}

/*
 * NAME:	os->read()
 * DESCRIPTION:	read bytes from an open descriptor
 */
long os_read(void **priv, void *buf, unsigned long len)
{
  PVOLINFO vol = (PVOLINFO)*priv;
  USHORT bytes;

  if (DosRead(vol->hf, buf, len, &bytes) != NO_ERROR)
    ERROR(EIO, 0);

  return bytes;

fail:
  return -1;
}

/*
 * NAME:	os->write()
 * DESCRIPTION:	write bytes to an open descriptor
 */
long os_write(void **priv, const void *buf, unsigned long len)
{
  PVOLINFO vol = (PVOLINFO)*priv;
  USHORT bytes;

  if (DosWrite(vol->hf, buf, len, &bytes) != NO_ERROR)
    ERROR(EIO, 0);

  return bytes;

fail:
  return -1;
}
